"""
general_template
"""

import logging

from peewee import *

from spider import ResourceRoot
from spider import Spider

spider = Spider()
logger = logging.getLogger('spider')
res = ResourceRoot('resources')
db = SqliteDatabase('db.sqlite')


class MyModel(Model):
    url = CharField()
    data = TextField()

    class Meta:
        database = db


if __name__ == '__main__':
    r = spider.get('https://www.baidu.com/')
    r.encoding = 'gb2313'
    result = r.css('title')[0]
    logger.info(result)
    res.close()



import re
import warnings

from lxml import etree
from lxml.html import HtmlElement

from .html2text import html2text


def elem_tostring(elem: HtmlElement, encoding='utf8'):
    """HTML元素转换成字符串"""
    warnings.warn('使用elem2text作为替代', DeprecationWarning)

    return elem2text(elem, encoding=encoding)


def elem2md(elem: HtmlElement, encoding='utf8'):
    return html2md(etree.tostring(elem, encoding=encoding))


def html2md(html: str):
    return html2text(html)


def elem2text_re(elem: HtmlElement, encoding='utf8'):
    warnings.warn('没完成', DeprecationWarning)

    html = etree.tostring(elem, encoding=encoding).decode()
    # text = re.findall(r'(<(?:.|\n)*?>)', html, re.I)
    text = re.sub(r'(<(?:.|\n)*?>)', '', html).replace('\n', '')
    print(text)
    return text


def elem2text(elem: HtmlElement, encoding='utf8'):
    """HTML元素转换成字符串"""
    elem_text_nodes = [str(e) for e in elem.xpath(".//text()") if e.isprintable()]
    beautiful_text = '\n'.join([e.strip() for e in elem_text_nodes])
    return beautiful_text


def md2text(md: str):
    warnings.warn('没完成', DeprecationWarning)



__author__ = 'Notnotype'

from typing import Callable, List

from peewee import *

HTTP = 1
HTTPS = 2
HTTP_HTTPS = 3
SOCK5 = 4
HTTP_SOCK5 = 5
HTTPS_SOCK5 = 6
HTTP_HTTPS_SOCK5 = 7


class ProxyPoolBase:
    def __init__(self):
        ...

    def get_proxies(self, count=1) -> List:
        ...

    def set_proxy(self, host, port, username=None, password=None, schema=HTTP):
        ...

    def del_proxy(self, host, port) -> int:
        ...

    def get_in_middlewares(self) -> List[Callable]:
        ...

    def add_in_middleware(self, middleware: Callable):
        ...

    def get_out_middlewares(self) -> List[Callable]:
        ...

    def add_out_middleware(self, middleware: Callable):
        ...


class SqliteProxy(Model):
    host = CharField()
    port = IntegerField()
    schema = IntegerField()
    username = CharField(null=True)
    password = CharField(null=True)


def default_in_middleware(model: SqliteProxy, host, port,
                          schema, username, password) -> SqliteProxy:
    model.host = host
    model.port = port
    model.schema = schema
    model.username = username
    model.password = password
    return model


def default_out_middleware(model: SqliteProxy, proxy: dict) -> dict:
    proxy = {}
    schema = model.schema
    _http = schema & 1
    _https = schema & 2
    _sock5 = schema & 4
    if _http:
        proxy['http'] = 'http://{username}{password}{host}:{port}'.format(
            username=model.username + ':' if model.username else '',
            password=model.password + '@' if model.password else '',
            host=model.host,
            port=model.port
        )
    if _https:
        proxy['https'] = 'https://{username}{password}{host}:{port}'.format(
            username=model.username + ':' if model.username else '',
            password=model.password + '@' if model.password else '',
            host=model.host,
            port=model.port
        )
    if _sock5:
        proxy['sock5'] = 'sock5://{username}{password}{host}:{port}'.format(
            username=model.username + ':' if model.username else '',
            password=model.password + '@' if model.password else '',
            host=model.host,
            port=model.port
        )
    return proxy


class SqliteProxyPool(ProxyPoolBase):

    def __init__(self, db: SqliteDatabase):
        super().__init__()
        self.in_middlewares: List[Callable] = [default_in_middleware]
        self.out_middlewares: List[Callable] = [default_out_middleware]
        # init database
        self.db = db
        SqliteProxy._meta.set_database(db)
        print('*******************************Ok')
        if not db.is_connection_usable():
            db.connect()
        if not db.table_exists(SqliteProxy):
            db.create_tables([SqliteProxy])

    def get_proxies(self, count=1) -> List:
        proxies = []
        query = SqliteProxy.select().order_by(self.db.random()).limit(count)
        for model in query:
            proxy = {}
            for middleware in self.out_middlewares:
                proxy = middleware(model, proxy)
            proxies.append(proxy)
        return proxies

    def set_proxy(self, host, port, schema=HTTP, username=None, password=None):
        query = SqliteProxy.select().where(SqliteProxy.host == host and SqliteProxy.port == port)
        if query:
            model = query[0]
        else:
            model = SqliteProxy()
        for middleware in self.in_middlewares:
            model = middleware(model, host, port, schema, username, password)
        model.save()

    def del_proxy(self, host, port):
        sql = SqliteProxy.delete().where(SqliteProxy.host == host and SqliteProxy.port == port)
        return sql.execute()

    def get_in_middlewares(self) -> List[Callable]:
        return self.in_middlewares

    def add_in_middleware(self, middleware: Callable):
        self.in_middlewares.append(middleware)

    def get_out_middlewares(self) -> List[Callable]:
        return self.in_middlewares

    def add_out_middleware(self, middleware: Callable):
        self.out_middlewares.append(middleware)



import datetime
import json
import logging
import os
import re
import shutil
import time
import uuid
from json import loads
from os.path import exists
from os.path import join
from time import localtime
from time import sleep
from types import MethodType
from typing import Union, IO, List, Callable
from urllib.parse import urljoin, quote

import requests
from lxml.etree import HTML
from lxml.html import HtmlElement

from spider.cache import SqliteCache
from .utils import general_response_pipeline, get_random_header, limit_text, random_sleeper


# 改变脚本的工作目录
# START_DIR = os.getcwd()
# FILE_DIR = os.path.dirname(os.path.abspath(__file__))
# os.chdir(FILE_DIR)


class FormatFilter(logging.Filter):

    def filter(self, record: logging.LogRecord) -> int:
        def getMessage(obj):
            msg = str(obj.msg)
            if obj.args:
                msg = msg.format(*obj.args)
            return msg

        # 使用`{`风格格式化
        record.getMessage = MethodType(getMessage, record)

        # context: dict = record.__getattribute__('context')
        # record.msg += '\n' + '\n'.join([f'{str(k)}: {str(v)}' for k, v in context.items()])

        return True


# todo 继承`StreamHandler`实现详细`log`与精简`log`
# todo 记录错误单独保存文件
def init_logger(log_dir='log', level=logging.DEBUG) -> logging.Logger:
    # os.chdir(START_DIR)
    if not exists(log_dir):
        os.makedirs(log_dir)
    file_handler = logging.FileHandler(f"{log_dir}/"
                                       f"{localtime().tm_year}-"
                                       f"{localtime().tm_mon}-"
                                       f"{localtime().tm_mday}--"
                                       f"{localtime().tm_hour}h-"
                                       f"{localtime().tm_min}m-"
                                       f"{localtime().tm_sec}s.log",
                                       encoding="utf-8")
    formatter = logging.Formatter('[{asctime}]'
                                  '[{levelname!s:5}]'
                                  '[{name!s:^6}]'
                                  '[{lineno!s:4}行]'
                                  '[{module}.{funcName}]\n'
                                  '{message!s}',
                                  style='{',
                                  datefmt='%Y-%m-%d %H:%M:%S')
    file_handler.setFormatter(formatter)

    _logger = logging.getLogger('spider')
    _console = logging.StreamHandler()
    _logger.setLevel(level)
    _console.setLevel(level)

    _logger.addHandler(file_handler)
    _console.setFormatter(formatter)

    _logger.addHandler(file_handler)
    _logger.addHandler(_console)
    _logger.addFilter(FormatFilter())
    # os.chdir(FILE_DIR)
    return _logger


logger = init_logger()


class ResourceBase:
    """ResourceRoot的无意义基类"""

    def serialize_as_folder(self, path):
        pass


# 考虑使用`property`
class ResourceRoot(ResourceBase):
    # todo 没有 close
    def scan(self):
        """扫描当前文件夹, 更新`list_dir`, `files`, `dirs`"""
        self.list_dir = list(map(lambda name: join(self.root_dir, name), os.listdir(self.root_dir)))

        file_names = list(filter(lambda name: os.path.isfile(name), self.list_dir))
        self.files = {filename: None for filename in file_names}
        dir_names = [_dir for _dir in filter(lambda name: os.path.isdir(name), self.list_dir)]
        self.dirs = {dir_name: ResourceRoot(dir_name) for dir_name in dir_names}

    def __init__(self, root_dir='resources', chuck=2048, mode='r+'):
        """把一个文件夹抽象成一个类,可以保存和读取资源文件

        :key root_dir: 默认为`resources`
        :key chuck: 默认读取写入的区块
        :key mode: 文件打开模式  默认`r+`
        """
        self.rel_root_dir = root_dir
        self.chuck = chuck
        self.root_dir = os.path.abspath(root_dir)
        self.__mode = mode
        # self.__encoding = encoding

        # These Attributes will be 初始化 before long
        self.list_dir = None
        self.files = None
        self.dirs = None

        if not exists(self.root_dir):
            os.makedirs(self.root_dir)
            logger.info('创建root_dir {}', self.root_dir)

        self.scan()

    @property
    def mode(self):
        return self.__mode

    @mode.setter
    def mode(self, value):
        self.__mode = value

    @staticmethod
    def __guess_file_mode(filename):
        """根据文件名, 或者 `url` 来猜测使用哪一种模式打开文件"""
        if filename.lower().endswith((
                '.jpg', '.png'
        )):
            mode = 'wb'
        else:
            mode = 'w'
        return mode

    def __getitem__(self, name: str) -> IO:
        name = join(self.root_dir, name)
        if name in self.files.keys():
            if self.files[name] is None or self.files[name].closed:
                mode = self.__guess_file_mode(name)
                if 'b' not in mode:
                    self.files[name] = open(name, mode=mode)
                else:
                    self.files[name] = open(name, mode=mode, encoding='w')
            return self.files[name]
        elif name in self.dirs.keys():
            return self.dirs[name]
        else:
            raise KeyError('不存在此文件')

    def __setitem__(self, filename: str, value: Union[str, IO, dict, ResourceBase, bytes]):
        """默认调用`self.save`"""
        filename = filename.format(
            now=datetime.datetime.now().strftime('%Y-%m-%d_%H:%M:%S'),
            uuid=uuid.uuid4().__str__().replace('-', ''),
            time=time.time().__str__().replace('.', '')
        )
        self.save(filename, value)

        # todo 低效率
        self.scan()

    def __contains__(self, name: str):
        return name in self.list_dir

    def __str__(self):
        # return '<ResourceRoot root_dir=\'{}\'>'.format(self.root_dir)
        return '<ResourceRoot {!r}>'.format(self.list_dir)

    def create_sub_root(self, name):
        name = join(self.root_dir, name)
        if not exists(name):
            os.mkdir(name)
        sub_root = ResourceRoot(name)
        self.dirs[name] = sub_root
        return sub_root

    def serialize_as_folder(self, path):
        """把自己的一个复制复制到一个位置"""
        shutil.copytree(self.rel_root_dir, os.path.realpath(path))

    def save(self, name, value: Union[str, IO, dict, ResourceBase, bytes], **kwargs):
        """传入文件名,和一个`流`, 或者`字符串`, 或者`ResourceRoot`, 保存文件后,流将被关闭
            自动覆盖

        :param name: 文件名你要保存在这个`ResourceRoot`下的
        :param value: 它可能是一个`str`对象, 或者`IO`流对象, 或者是一个`dict`字典,如果传入`dict`则会被转换成`json`文本保存
        """
        # 如果是 `ResourceBase` 类
        if isinstance(value, ResourceBase):
            value.serialize_as_folder(join(self.root_dir, name))
            logger.debug('保存目录成功[{}]', join(self.root_dir, name))
        else:  # 是字符串或流
            path = join(self.root_dir, name)
            encoding = kwargs.get('encoding', None)
            # 如果是字典, 则把字典转换为字符串
            if isinstance(value, dict):
                f = open(path, mode='w', encoding=encoding)
                json.dump(value, f, indent=4, ensure_ascii=False)
                # 如果是字符串
            elif isinstance(value, str):
                f = open(path, mode='w', encoding=encoding)
                f.write(value)
            elif isinstance(value, bytes):
                f = open(path, 'wb')
                f.write(value)
            # 如果是流
            elif isinstance(value, IO):
                f = open(path, 'wb')
                f.write(value.read())
            else:
                raise RuntimeError('不支持传入此类型{}'.format(type(value)))
            logger.debug('保存文件成功[{}]', join(self.root_dir, name))

    def close(self):
        ...


# 抽象来使用 `pipeline`
class Spider:
    # todo 针对每次请求不同的`header`来重新加载缓存
    # todo 增加字段`data`,存`post`字段
    # todo 增加字段`header`, 存header

    # 强制使用缓存
    FORCE_CACHE = 2
    # 运行使用缓存
    ENABLE_CACHE = 1
    # 不使用缓存
    DISABLE_CACHE = 0

    # done `lxml`
    class Response:
        def __init__(self, response: requests.Response):
            self.response = response
            self.status_code = response.status_code
            self.__html = None
            self.__json = None
            self.__title = None

        @property
        def history(self):
            return self.response.history

        @property
        def content(self):
            return self.response.content

        @property
        def text(self):
            return self.response.text

        @property
        def encoding(self):
            return self.response.encoding

        @encoding.setter
        def encoding(self, encoding):
            self.response.encoding = encoding

        @property
        def url(self):
            return self.response.url

        @property
        def html(self) -> HtmlElement:
            if self.__html is None:
                self.__html = HTML(self.text)
            return self.__html

        def xpath(self, exp) -> List[HtmlElement]:
            return self.html.xpath(exp)

        def css(self, css) -> List[HtmlElement]:
            return self.html.cssselect(css)

        @property
        def json(self, *args, slices=None, **kwargs) -> dict:
            """
            :keyword slices: (start, end)字符串分片后在进行解码
            """
            if not self.__json:
                try:
                    if slices:
                        self.__json = loads(self.text[slices[0]: slices[1]], *args, **kwargs)
                    else:
                        self.__json = loads(self.text, *args, **kwargs)
                except json.JSONDecodeError:
                    logger.error("Json解码错误{}", self.text)
                    raise
            return self.__json

        def search(self, pattern, flags=0):
            return re.search(pattern, self.text, flags=flags)

        @property
        def ok(self):
            return self.response.ok

        @property
        def title(self):
            if not self.__title:
                result = self.__title = self.xpath('//title/text()')
                if result:
                    self.__title = result[0]
                else:
                    self.__title = "Error"
            return self.__title

        def __repr__(self):
            return '<Response[{}] [{}]>'.format(
                self.response.status_code,
                self.title
            )

        def __str__(self):
            return self.__repr__()

    def __init__(self):
        self.headers_generator = get_random_header
        # self.cache = JsonCache()
        self.cache = SqliteCache()
        self.session = requests.session()
        self.sleeper = random_sleeper(5, 10)
        # todo change to pipelines
        self.response_pipeline = general_response_pipeline
        self.request_pipeline = None
        # todo proxy_generator
        self.proxy_generator = None
        self.__encoding = None
        self.update_headers()

    def set_sleeper(self, sleeper: Callable):
        if not isinstance(sleeper, Callable):
            raise RuntimeError('参数必须是Callable')
        self.sleeper = sleeper

    def set_header_generator(self, header_generator):
        # if not isinstance(header_generator, Callable):
        #     raise RuntimeError('参数必须是Callable')
        self.headers_generator = header_generator

    def set_response_pipeline(self, response_pipeline):
        # if not isinstance(response_pipeline, Callable):
        #     raise RuntimeError('参数必须是Callable')
        self.response_pipeline = response_pipeline

    def set_request_pipeline(self, request_pipeline):
        # if not isinstance(request_pipeline, Callable):
        #     raise RuntimeError('参数必须是Callable')
        self.request_pipeline = request_pipeline

    def set_proxy_generator(self, proxy_generator):
        # if not isinstance(proxy_generator, Callable):
        #     raise RuntimeError('参数必须是Callable')
        self.proxy_generator = proxy_generator

    @property
    def encoding(self):
        return self.__encoding

    @encoding.setter
    def encoding(self, encoding):
        self.__encoding = encoding

    def set_cookie(self, cookie: str):
        """设置cookie
        :param cookie: 一个 `cookie` 字符串
        """
        cookie = {'Cookie': cookie}
        self.session.cookies = requests.sessions.cookiejar_from_dict(
            cookie,
            cookiejar=None,
            overwrite=True)

    def lunch(self, handle: [requests.get, requests.post], *args, **kwargs) -> \
            Union[Response, requests.Response, object]:
        """

        :param handle 处理请求的一个函数
        :param kwargs: 包含`requests`库所有选项
        :keyword alive_time: Union[datetime, int]缓存存活日期
        :keyword cache: 是否使用缓存
        :keyword sep_time: 间隔时间

        :return: Union[Response, requests.Response, object]
        """
        # 获取`alive_time`, `url`参数
        kwargs.setdefault('alive_time', datetime.datetime.now() + datetime.timedelta(days=3))
        kwargs.setdefault('cache', True)
        kwargs.setdefault('timeout', (5, 20))
        kwargs.setdefault('retry', 3)
        alive_time = kwargs.pop('alive_time')
        sep_time = kwargs.get('sep_time', self.sleeper)
        cache_enable = kwargs.pop('cache')
        retry = kwargs.pop('retry')

        url = ''
        for each in args:
            url = urljoin(url, each)
        if '://' not in url:
            url = 'http://' + url
            logger.debug('url没有添加协议, 使用[http]协议代替')

        is_force_cache = cache_enable == self.FORCE_CACHE

        # 如果 `is_force_cache` is True 则, 不论缓存是否过期, 都从缓存加载
        if self.cache.is_cached(url, ignore_date=is_force_cache) and cache_enable:
            logger.debug('从缓存: {} <- {}', limit_text(url, 200), self.cache.table_name)
            return self.cache.from_cache(url, force=is_force_cache)

        if cache_enable == Spider.DISABLE_CACHE:
            # 如果禁用这个url的缓存, 则将之从缓存文件删除
            self.cache.clear_cache(url)

        class HTTPBadCodeError(RuntimeError):
            ...

        # 发送请求
        # 显示logger文案
        if 'params' in kwargs:
            params = kwargs.get('params')
            tmp = ['{}={}'.format(k, quote(v)) for k, v in params.items()]
            logger.info('请求: {}', limit_text(url + '?' + '&'.join(tmp), 200))
        else:
            logger.info('请求: {}', limit_text(url, 200))

        retry = retry
        while retry:
            try:
                # 间隔时间
                response = self.Response(handle(url, **kwargs))
                if isinstance(sep_time, Callable):
                    sep_time()
                else:
                    sleep(sep_time)
                if len(response.response.history) >= 1:
                    history = [each.url for each in response.response.history]
                    history.append(response.url)
                    logger.debug('页面重定向: {}',
                                 '->'.join(['[{}]'.format(i) for i in history]))
                if response.response.ok:
                    if cache_enable == Spider.ENABLE_CACHE:
                        self.cache.cache(response.url, response, alive_time)
                else:
                    raise HTTPBadCodeError(f'坏HTTP响应', response)
                return response
            except requests.Timeout as e:
                if retry == 1:
                    raise e
                logger.debug('超时,重试---{}'.format(str(4 - retry)))
            except requests.RequestException as e:
                if retry == 1:
                    logger.error('取消重试---{}'.format(str(4 - retry)))
                    raise e
                logger.error('HTTP报错---{}'.format(str(4 - retry)))
                # todo 对于失败的`url`保存到另一个`log`文件
            except HTTPBadCodeError as e:
                if retry == 1:
                    logger.info('坏HTTP响应, 取消重试({})'.format(e.args[1].status_code))
                    return e.args[1]
                logger.debug('坏HTTP响应({})---{}', e.args[1].status_code, 4 - retry)
            finally:
                retry -= 1

    def get(self, *args, **kwargs) -> [Response, requests.Response, object]:
        """获取网页

        :param args: (元组)`url`的各个路径:
        :param kwargs: 包含`requests`库所有选项
        :keyword alive_time: Union[datetime, int]缓存存活日期
        :keyword cache: 是否使用缓存
        :keyword sep_time: 间隔时间

        :return: Union[Response, requests.Response, object]
        """
        # 获取`alive_time`, `url`参数
        resp = self.lunch(self.session.get, *args, **kwargs)
        if self.response_pipeline:
            resp = self.response_pipeline(self, resp)
        return resp

    def post(self, *args, **kwargs) -> Response:
        """获取网页

        :param args: (元组)`url`的各个路径:
        :param kwargs: 包含`requests`库所有选项
        :keyword alive_time: Union[datetime, int]缓存存活日期
        :keyword cache: 是否使用缓存
        :keyword sep_time: 间隔时间

        :return: Union[Response, requests.Response, object]
        """
        resp = self.lunch(self.session.post, *args, **kwargs)
        if self.response_pipeline:
            resp = self.response_pipeline(self, resp)
        return resp

    def update_headers(self):
        """调用`self.headers_generator`来更新头"""
        if self.headers_generator:
            self.session.headers.update(self.headers_generator())

    def close(self):
        self.session.close()


local = {}


def get_spider():
    if 'spider' not in local:
        local['spider'] = Spider()
    return local['spider']


def gs():
    """qs() == get_spider()"""
    return get_spider()


def get(*args, **kwargs):
    gs().get(*args, **kwargs)


def post(*args, **kwargs):
    gs().post(*args, **kwargs)


def close():
    gs().close()



import json
import traceback
from typing import IO
from typing import Iterable


class Meta(type):
    def __new__(mcs, name, base, attr):
        mappings = {}
        for k, v in attr.items():
            if isinstance(v, Field):
                mappings[k] = v
        for k in mappings:
            attr.pop(k)
        attr['__mappings__'] = mappings
        return type.__new__(mcs, name, base, attr)


class Field:
    def __init__(self, name=None, value=None):
        self.name = name
        self.value = value

    def __repr__(self):
        return '{}'.format(self.__class__.__name__)


class JsonItemMixin:
    """实现`dump`方法, 把一个`Item`类转化为`json`"""

    def __init__(self):
        ...

    def dump(self):
        dumping = {}
        for k, v in self.__mappings__.items():
            print('dumping: ', k, v)
            dumping[k] = v
        return dumping


class FileItemMixin:
    """实现`save`, `flush` 方法, 保存这个`Item`为一个文件"""

    def __init__(self):
        ...

    def save(self):
        ...


class Item(metaclass=Meta):
    def __init__(self):
        self.obj = {}

    def keys(self) -> Iterable:
        return self.obj.keys()

    def items(self) -> (Iterable, Iterable):
        return self.items()

    def __len__(self) -> int:
        return self.obj.__len__()

    def __setitem__(self, k, v):
        return self.obj.__setitem__(k, v)

    def __getitem__(self, k):
        return self.obj.__getitem__(k)

    def __delitem__(self, v):
        return self.obj.__delitem__(v)

    def __str__(self) -> str:
        return self.obj.__str__()


class JsonItem(Item, JsonItemMixin):
    def __init__(self):
        super(JsonItem, self).__init__()


class JsonFile(Item):

    def __init__(self, file: IO, obj: dict = None, indent=4):
        """Json与文件同步序列化
        注意不要使用此构造方法,应当使用工厂方法

        :param file: 一个文件流
        :param obj: 一个字典
        :param indent: tab的长度
        """
        super().__init__()
        if not file.writable():
            raise IOError('文件不可写<{}>'.format(file.name))
        if not file.seekable():
            raise IOError('文件不可查<{}>'.format(file.name))
        if not file.readable():
            raise IOError('文件不可读<{}>'.format(file.name))
        self.f = file
        self.obj = obj
        self.indent = indent

    @classmethod
    def from_newfile(cls, filename, mode='w', encoding='utf8'):
        f = open(filename, mode=mode, encoding=encoding)
        return cls(f, {})

    @classmethod
    def from_filename(cls, filename, mode='r+', encoding='utf8'):
        f = open(filename, mode=mode, encoding=encoding)
        c = cls(f)
        c.load(f)
        return c

    @classmethod
    def from_streaming(cls, streaming):
        c = cls(streaming, {})
        return c

    def load(self, streaming: IO):
        self.obj = json.load(streaming)

    def dump(self):
        self.f.seek(0)
        json.dump(self.obj, self.f, indent=4, ensure_ascii=False)

    def close(self):
        self.dump()
        self.f.close()


# #################################test Mapping###########################

class MyItem(Item):
    def __init__(self):
        super().__init__()

    field = Field()


class MyJsonItem(JsonItem):
    field = Field()


class Test:
    def test_json_item(self):
        mji = MyJsonItem()
        mji.dump()

    def test_mapping(self):
        it = MyItem()
        mp = it.__mappings__
        print('mp={}'.format(mp))

    def test_json2(self):
        jf2 = JsonFile.from_filename('resource/testjson2.json')
        jf2['title'] = 'changeadsfsdafdsd testjson2'
        jf2.close()

    # noinspection PyBroadException
    def start_test(self):
        method_names = [method for method in self.__dir__() if method.startswith('test')]
        for method_name in method_names:
            try:
                print('======test: [{}]======'.format(method_name))
                getattr(self, method_name)()
            except:
                print('======Exception=======')
                traceback.print_exc()


if __name__ == '__main__':
    test = Test()
    test.start_test()



"""
生成模板
"""

from .spider import ResourceRoot


def render_template(template: str, data: dict):
    return template.format(**data)


def init(dirname: str, name: str, template_style: int):
    """初始化项目"""
    template_root = ResourceRoot('templates')
    dir_root = ResourceRoot(dirname)

    template_root.close()
    dir_root.close()



import datetime
import json
import logging
import os
import pickle
import uuid
from json import load, dump
from math import ceil
from os.path import exists
from os.path import join
from typing import Union

from peewee import *

from .utils import limit_text, get_sqlite_db

logger = logging.getLogger('spider')
db = get_sqlite_db()


class CacheBase:
    def __init__(self):
        self.cache_size = 10000
        # "filename": str(uuid.uuid4()),
        # "typing": str(obj.__class__),
        # "alive_time": alive_time.strftime('%Y-%m-%d %H:%M:%S')

    def is_cached(self, name: str, ignore_date=False) -> bool:
        ...

    def from_cache(self, name: str, force=False) -> object:
        ...

    def cache(self, name: str, obj, alive_time: Union[datetime.datetime, int]) -> bool:
        ...

    def save(self):
        ...

    def clear_cache(self, name: str):
        ...

    def clear_all(self):
        ...


class JsonCache(CacheBase):
    """
    __init__
    cache
    from_cache
    is_cached
    close
    """

    def __init__(self, cache_dir='__pycache__'):
        super().__init__()
        self.__cache_dir = cache_dir
        self.__cache_json_name = 'cache.json'
        self.cache_size = 10000
        if not exists(cache_dir):
            os.makedirs(cache_dir)
            logger.debug('生成文件缓存路径{}', os.path.realpath(cache_dir))
        # 是否存在`cache.json`,没有则生成
        if not exists(join(cache_dir, self.__cache_json_name)):
            try:
                with open(join(self.__cache_dir, self.__cache_json_name), 'a+', encoding='utf8') as f:
                    dump({"cached_files": {}}, f, indent=4, ensure_ascii=False)
            except IOError as e:
                logger.error('IO错误{}', join(self.__cache_dir, self.__cache_json_name))
                raise e
        # 打开缓存清单文件
        try:
            with open(join(cache_dir, self.__cache_json_name), 'r', encoding='utf8') as f:
                self.__cache_json: dict = load(f)
        except IOError as e:
            logger.exception(e)
            logger.error('IO错误{}', join(cache_dir, self.__cache_json_name))
            raise e
        except json.JSONDecodeError as e:
            logger.error('Json解码错误{}', join(cache_dir, self.__cache_json_name))
            raise e
        except Exception as e:
            logger.error('未知错误在{}', join(cache_dir, self.__cache_json_name))
            raise e

    def is_cached(self, name: str, ignore_date=False) -> bool:
        if name in self.__cache_json['cached_files'].keys():
            item = self.__cache_json['cached_files'][name]
            alive_time: datetime.datetime = datetime.datetime.strptime(item['alive_time'], '%Y-%m-%d %H:%M:%S')
            if alive_time > datetime.datetime.now() or ignore_date:
                return True
            else:
                logger.debug('存活时间已过,重新缓存')
                return False
        else:
            return False

    def from_cache(self, name: str, force=False) -> object:
        """

        :param name:
        :param force: 忽略时间过期, 强制从缓存读取
        :return:
        """
        if self.is_cached(name, ignore_date=force):
            item = self.__cache_json['cached_files'][name]
            filename = item['filename']
            with open(join(self.__cache_dir, filename), 'rb') as f:
                return pickle.load(f)
        else:
            return None

    def cache(self, name: str, obj, alive_time: Union[datetime.datetime, int]) -> bool:
        if isinstance(alive_time, int):
            alive_time = datetime.datetime.now() + datetime.timedelta(days=alive_time)
        self.__cache_json['cached_files'][name] = {
            "filename": str(uuid.uuid4()),
            "typing": str(obj.__class__),
            "alive_time": alive_time.strftime('%Y-%m-%d %H:%M:%S')
        }
        item = self.__cache_json['cached_files'][name]
        filename = item['filename']
        try:
            with open(join(self.__cache_dir, filename), 'wb') as f:
                pickle.dump(obj, f)
                self.save()
        except IOError:
            logger.error('IO错误: {}'.format(filename))
            return False
        else:
            logger.debug('缓存: {} -> {}'.format(limit_text(name, 200), filename))
            return True

    # 关闭保存文件
    def save(self):
        try:
            with open(join(self.__cache_dir, self.__cache_json_name), 'r+', encoding='utf8') as f:
                dump(self.__cache_json, f, indent=4, ensure_ascii=False)
        except IOError as e:
            logger.error('IO错误: {}', join(self.__cache_dir, self.__cache_json_name))
            raise e

    def clear_cache(self, name: str):
        if name in self.__cache_json['cached_files']:
            item = self.__cache_json['cached_files'][name]
            filename = item['filename']
            os.remove(join(self.__cache_dir, filename))
            del self.__cache_json['cached_files'][name]
            logger.debug('删除缓存[{}]'.format(name))

    def clear_all(self):
        logger.debug('删除缓存[{}]个'.format(len(self.__cache_json)))
        self.__cache_json = {"cached_files": {}}


class SqliteCacheData(Model):
    url = CharField()
    typing = CharField()
    alive_time = DateTimeField()
    data = BitField()

    class Meta:
        database = db
        table_name = 'lazy_spider_cache'


class SqliteCache(CacheBase):

    def __init__(self):
        super().__init__()
        if not db.is_connection_usable():
            db.connect()
        if not db.table_exists(SqliteCacheData):
            db.create_tables([SqliteCacheData])
        self.table_name = str(SqliteCacheData)
        self.cache_size = 10000
        self.remain_percent = 0.75

    def is_cached(self, name: str, ignore_date=False) -> bool:
        query = SqliteCacheData.select().where(SqliteCacheData.url == name)
        if query:
            item = query[0]
            alive_time: datetime.date = item.alive_time
            if alive_time > datetime.datetime.now() or ignore_date:
                return True
            else:
                logger.debug('存活时间已过,重新缓存')
                self.clear_cache(name)
                return False
        else:
            return False

    def from_cache(self, name: str, force=False) -> object:
        """

        :param name:
        :param force: 忽略时间过期, 强制从缓存读取
        :return:
        """
        if self.is_cached(name, ignore_date=force):
            item = SqliteCacheData.select().where(SqliteCacheData.url == name)[0]
            return pickle.loads(item.data)
        else:
            return None

    def cache(self, name: str, obj, alive_time: Union[datetime.datetime, int]) -> bool:
        # 清除过多的缓存
        self.clear_more_caches()
        if isinstance(alive_time, int):
            alive_time = datetime.datetime.now() + datetime.timedelta(days=alive_time)
        query = SqliteCacheData.select().where(SqliteCacheData.url == name)
        # 查询
        if not query:
            item = SqliteCacheData()
        else:
            item = query[0]
        # 填充数据
        item.url = name
        item.typing = str(obj.__class__)
        item.alive_time = alive_time.strftime('%Y-%m-%d %H:%M:%S')
        item.data = pickle.dumps(obj)
        logger.debug('缓存: {} -> {}'.format(limit_text(name, 200),
                                           str(SqliteCacheData)))
        return item.save()

    def save(self):
        super().save()

    def clear_more_caches(self):
        if SqliteCacheData.select().count() > self.cache_size:
            logger.debug('删除缓存中......')
            will_del = SqliteCacheData.select() \
                .limit(ceil(self.remain_percent * self.cache_size)) \
                .order_by(SqliteCacheData.alive_time)
            # print('will_del len', will_del.count())
            # del_count: ModelObjectCursorWrapper = will_del.execute()
            for each in will_del:
                each.delete_instance()
            logger.info('缓存过多, 最大缓存大小[{}], 删除缓存[{}]个, 剩下缓存[{}]个'.format(
                self.cache_size,
                will_del.__len__(),
                SqliteCacheData.select().count()
            ))

    def clear_cache(self, name: str):
        result = SqliteCacheData.delete().where(SqliteCacheData.url == name).execute()
        if result:
            logger.debug('删除缓存[{}]'.format(name))
        return result

    def clear_all(self):
        sql = SqliteCacheData.delete()
        result = sql.execute()
        logger.debug('删除缓存[{}]个'.format(result))
        return



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['spider',
 'spider.gerneric',
 'spider.templates',
 'spider.utils',
 'spider.utils.parse']

package_data = \
{'': ['*']}

install_requires = \
['fake-useragent>=0.1.11,<0.2.0',
 'lxml>=4.6.2,<5.0.0',
 'peewee>=3.14.0,<4.0.0',
 'requests>=2.25.1,<3.0.0']

setup_kwargs = {
    'name': 'lazy-spider',
    'version': '0.1.5',
    'description': 'A lazy spider tools which intergrate lxml, requests, peewee......',
    'long_description': None,
    'author': 'notnotype',
    'author_email': '2056369669@qq.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging
import random
from time import sleep
from typing import Iterable

from fake_useragent import UserAgent
from peewee import SqliteDatabase
from requests import Request

logger = logging.getLogger('spider')


def general_request_pipeline(spider, response) -> Request:
    ...


def general_response_pipeline(spider, response):
    return response


def get_random_header():
    """返回一个随机的头"""
    return {'User-Agent': str(UserAgent().random)}


def limit_text(s: str, max_len, replace='...'):
    """文本太长自动打省略号"""
    s_len = len(s)
    replace_len = len(replace)
    if s_len + replace_len > max_len:
        return s[:int(max_len / 2)] + replace + s[-int(max_len / 2):]
    else:
        return s


def good_dirname(string: str) -> str:
    """一个好目录名字"""
    string.replace('\n', '').replace('\t', '').replace(' ', '')
    string = limit_text(string, 60, '___')
    return string


def get_sqlite_db(db_name='db.sqlite'):
    db = SqliteDatabase(db_name)
    logger.info('获取数据库[{}]'.format(db_name))
    return db


def random_sleeper(a, b):
    def sleeper():
        sleep(random.randint(a, b))

    return sleeper


class ProxyGeneratorBase:
    def __call__(self, *args, **kwargs) -> Iterable:
        ...
